# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 16:10:15 2023

@author: Antto
"""

import numpy as np
import pandas as pd
from patsy import dmatrix
from scipy.optimize import minimize

np.seterr(all='print')

class DemandEstimation():
    def __init__(self, 
                 data: pd.DataFrame, # dataset
                 X1_formulation: str, # linear variables
                 X2_formulation: str, # non-linear variables
                 Z_d: str, # demand side instruments
                 init_sigma: np.array, # initial values for sigma
                 I: int = 1000, # number of draws of heterogeneity
                 seed: int = 1000): # random seed
        
        self.data = data # data
        self.rng = np.random.default_rng(seed=seed) # rng
        self.I = I # draws of heterogeneity
        
        self.market_indices = {market: np.where(self.data['market_ids'] == market)[0] for market in self.data['market_ids'].unique()}
        self.product_indices = {product: np.where(self.data['product_ids'] == product)[0] for product in self.data['product_ids'].unique()}
        
        self.N = self.data.shape[0] # number of observations
        self.T = self.data['market_ids'].nunique() # number of markets
        self.J = self.data['product_ids'].nunique() # number of products
        
        self.X1 = dmatrix(X1_formulation, data, return_type='dataframe').to_numpy() # linear variables
        self.X1_labels = dmatrix(X1_formulation, data).design_info.column_names # labels of linear variables
        
        if X2_formulation is not None:
            self.X2 = dmatrix(X2_formulation, data, return_type='dataframe').to_numpy() # non-linear variables
            self.X2_labels = dmatrix(X2_formulation, data).design_info.column_names # labels of non-linear variables
            self.sigma = init_sigma.copy() # starting guess for sigma
            self.v = self.rng.standard_normal(size=(self.I, self.X2.shape[1])) # draw heterogeneity
        else:
            self.X2 = np.empty(shape=(0, 0))
            self.X2_labels = None
            self.sigma = np.empty(shape=(0, ))
        
        price_pos = self.X1_labels.index('prices') # index of price in linear variables
        self.X1_endog = self.X1[:, price_pos] # endogenous linear variables
        self.X1_exog = np.delete(self.X1, obj=price_pos, axis=1) # exogenous linear variables
        
        self.Z_d = Z_d # demand-side instruments
        if Z_d is None: 
            self.__calc_hausman_instruments() # if no instruments, calculate hausman instruments
        self.Z = np.concatenate((self.Z_d, self.X1_exog), axis=1) # add exogenous vars to demand side instruments
        self.M = self.Z.shape[1] # number of moment conditions
        
        self.S = (self.Z.T @ self.Z)/self.N # moment covariances
        self.W = np.linalg.inv(self.S) # weighting matrix
        
        self.delta = np.zeros(shape=(self.N, 1)) # init mean-utilities
        
        self.beta_len = self.X1.shape[1] # number of exogenous linear parameters
        self.sigma_len = self.X2.shape[1] # number of non-linear parameters
        self.params_len = self.beta_len + self.sigma_len # number of parameters
        
        self.beta = None # init beta
        self.beta_se = None # init standard errors
        self.sigma_se = None # init standard errors
        
        self.shares = data['shares'].to_numpy()[:, np.newaxis] # market shares
            
        self.threshold = 1e-10 # threshold for contraction mapping stop-condition
        
        self.res = None # optimizer result class
        
        
        
    def __logit_probs(self, x):
        """
        Calculates discrete choice probabilities / shares when assumed
        Type 1 EV error terms (Logit).
        
        All elements in x are considered to be in the choice set and
        outside good is automatically included. Uses shifting, see: 
        https://stackoverflow.com/questions/42599498/numerically-stable-softmax

        Args:
            self, x
        Returns:
            np.array: probabilities
        Raises:
            -
        """
        outside_good = np.zeros(shape=(1, self.I)) # outside good
        x = np.concatenate((x, outside_good), axis=0) # add outside good to x
        x_max = x.max(axis=0) # largest x for each v
        exp_x = np.exp(x - x_max) # exp of shifted x
        probs = exp_x / np.sum(exp_x, axis=0) # calc probs
        probs = probs[:-1,:] # drop the outside good
        return probs
    
    
    def __mean_utilities(self, sigma):
        """
        Calculates mean utilities (delta) given a guess of non-linear 
        parameters (sigma).
        
        Args:
            self, sigma
        Returns:
            np.array: delta
        Raises:
            -
        """
        delta = self.delta # init mean component delta
        mu = self.X2 @ (self.v * sigma).T # draw individual-specific component mu
        for market, indices in self.market_indices.items(): # loop over markets
            shares_t = self.shares[indices, :] # select shares of market t
            delta_t = delta[indices, :] # select deltas of market t
            mu_t = mu[indices, :] # select mus of market t
            while True: # loop until convergence
                shares_hat_t = self.__logit_probs(delta_t + mu_t) # calc choice probabilities
                shares_hat_t = np.mean(shares_hat_t, axis=1)[:, np.newaxis] # average choice probabilities over v
                diff = np.log(shares_t) - np.log(shares_hat_t) # difference not captured by deltas
                if (np.absolute(diff) < self.threshold).all():
                    break # stop if convergence
                else:
                    delta_t = delta_t + diff # update delta
            delta[indices, :] = delta_t # insert delta_t into full vector of deltas
        self.delta = delta # set this delta to be starting value for next call
        return delta # return mean utilities

    
    def __iv_gmm(self, delta):
        """
        Solves IV-GMM problem for linear parameters given mean-utilities (delta).
        See: https://en.wikipedia.org/wiki/Instrumental_variables_estimation#Estimation

        Args:
            self, delta
        Returns:
            np.array: beta
        Raises:
            -
        """
        P_z = self.Z @ self.W @ self.Z.T # projection matrix
        cov_X = self.X1.T @ P_z @ self.X1 # variance matrix of X
        cov_Y = self.X1.T @ P_z @ delta # covariance matrix of X and Y
        beta = np.linalg.inv(cov_X) @ cov_Y # iv-gmm estimates for linear params
        return beta
    

    def __gmm_objective_function(self, sigma):
        """
        Given a guess of sigma, solves the value of GMM objective function.

        Args:
            self, sigma
        Returns:
            np.array: GMM error
        Raises:
            -
        """
        delta = self.__mean_utilities(sigma) # get mean utilities with contraction mapping
        beta = self.__iv_gmm(delta) # estimate linear params using iv gmm
        xi = delta - self.X1 @ beta # structural errors
        g_hat = self.Z.T @ xi / self.N # orthogonality conditions by observation
        gmm_error = g_hat.T @ self.W @ g_hat # gmm objective function
        print("GMM objective:", self.N * gmm_error[0][0])
        print("Sigma:", sigma)
        return self.N * gmm_error[0][0]
    
        
    def estimate(self, twostep):
        """
        Estimates the initialized model.

        Args:
            self
        Returns:
            np.array: SciPy results class
        Raises:
            -
        """
        init_sigma = self.sigma # starting values of sigma
        bounds = [(0, None) for sigma in init_sigma] # bounds for sigmas
        res = minimize(self.__gmm_objective_function, 
                       init_sigma, 
                       method='L-BFGS-B',
                       bounds=bounds,
                       options={'disp':True,
                                'gtol':1e-6}) # estimate the model
        self.res = res # save optimizer result class
        self.sigma = res.x # save sigmas
        delta = self.__mean_utilities(self.sigma) # get mean utilities
        self.beta = self.__iv_gmm(delta)[:, 0] # solve linear parameters
        if twostep: # if two-step estimation, update weighting matrix and repeat
            self.S = self.__moment_covariances() # moment covariances
            self.W = np.linalg.inv(self.S) # update weighting matrix
            res = minimize(self.__gmm_objective_function, 
                           self.sigma,
                           method='L-BFGS-B',
                           bounds=bounds, 
                           options={'disp':True,
                                    'gtol':1e-6}) # estimate the model
            self.res = res # save optimizer result class
            self.sigma = res.x # save sigmas
            delta = self.__mean_utilities(self.sigma) # get mean utilities
            self.beta = self.__iv_gmm(delta)[:, 0] # solve linear parameters
        self.__standard_errors()
        return res
    
    
    def __moment_jacobian(self, params):
        """
        Calculates jacobian with respect to GMM moment conditions.

        Args:
            self
        Returns:
            np.array: moment condition jacobian
        Raises:
            -
        """
        def xi(params):
            beta, sigma = np.split(params, [self.beta_len])
            delta = self.__mean_utilities(sigma)
            xi = delta - self.X1 @ beta[:, np.newaxis]
            return xi
        dxi = np.empty(shape=(self.N, self.params_len))
        for p in range(self.params_len): # loop over parameters
            dx = np.zeros(shape=params.shape) # initialize zero vector
            dx[p] = 1e-6 # set a change of dx to parameter n 
            dy = xi(params + dx/2) - xi(params - dx/2) # calc change in xi
            dxi[:, p] = (dy / dx[p])[:,0] # insert to dxi
        jacobian = self.Z.T @ dxi / self.N # multiply with instruments and average
        return jacobian
    
    
    def __moment_covariances(self):
        """
        Returns covatiance matrix of GMM moment conditions.

        Args:
            self
        Returns:
            np.array: covariance matrix of GMM errors
        Raises:
            -
        """
        xi = self.delta - self.X1 @ self.beta[:, np.newaxis] # structural errors
        g_hat = self.Z * xi # orthogonality condition
        cov = g_hat.T @ g_hat / self.N
        return cov
    
    def __standard_errors(self):
        """
        Calculates standard errors of parameters.
        See https://pyblp.readthedocs.io/en/stable/background.html#standard-errors

        Args:
            self
        Returns:
            -
        Raises:
            -
        """
        unadjusted = True
        params = np.concatenate((self.beta, self.sigma), axis=0)
        G_hat = self.__moment_jacobian(params)
        cov = np.linalg.inv(G_hat.T @ self.W @ G_hat)
        if not unadjusted:
            cov = cov @ G_hat.T @ self.W @ self.S @ self.W @ G_hat @ cov
        se = np.sqrt(np.diagonal(cov)) / self.N
        self.beta_se, self.sigma_se = np.split(se, [self.beta_len])
    
    def __demand_jacobian(self, var, market_id):
        """
        Calculates jacobian of market shares with respect to specified variable.
        Calculates jacobian for one market.

        Args:
            self
        Returns:
            np.array: moment condition jacobian
        Raises:
            -
        """
        indices = self.market_indices[market_id]
        J = len(indices)
        X1 = self.X1[indices, :]
        X2 = self.X2[indices, :]
        jacobian = np.empty(shape=(J, J))
        dx = 1e-6
        for j in range(J):
            dX1 = np.zeros(shape=(X1.shape)) # initialize zero vector
            if var in self.X1_labels:
                var_pos = self.X1_labels.index(var)
                dX1[j, var_pos] = dx # set a change of dx
            dX2 = np.zeros(shape=(X2.shape)) # initialize zero vector
            if var in self.X2_labels:
                var_pos = self.X2_labels.index(var)
                dX2[j, var_pos] = dx # set a change of dx
            dy = self.get_market_shares(X1+dX1/2, X2+dX2/2) - self.get_market_shares(X1-dX1/2, X2-dX2/2)
            jacobian[:, j] = dy / dx
        return jacobian
    

        
    def get_market_shares(self, X1, X2):
        """
        Returns market shares given the observable characterics. Assumes all 
        observations to be in the same markets.

        Args:
            self, X1, X2
        Returns:
            np.array: market shares
        Raises:
            -
        """
        delta = X1 @ self.beta # calc delta
        mu = X2 @ (self.v * self.sigma).T # draw individual-specific component mu
        u = delta[:, np.newaxis] + mu # sum up
        shares = self.__logit_probs(u) # calc shares with logit
        shares = np.mean(shares, axis=1) # average over v
        return shares # return shares
    
                
    def marginal_costs(self, ownership):
        """
        Calculates marginal costs. Takes in ownership dictionary which has an 
        ownership matrix for each market (allowing for different matrices or
        choice sets over markets).
        
        Args:
            self, ownership
        Returns:
            np.array: marginal costs
        Raises:
            -
        """
        prices = self.data['prices'].to_numpy() # get prices
        shares = self.shares # get shares
        mc = np.empty(shape=(self.N, 1)) # initalize marginal cost vector
        for market, indices in self.market_indices.items(): # loop over markets
            price_t = prices[indices] # select prices in market t
            jac = self.__demand_jacobian(var='prices', market_id=market) # get demand jacobian wrt price
            omega = ownership[market] * (-jac) # calc omega
            mc[indices] = price_t[:, np.newaxis] - np.linalg.inv(omega) @ shares[indices, :] # calc marginal costs
        return mc
    
    
    
    ### Everything after this is WIP ###
    
    
    def __gmm_gradient(self, sigma, beta, g_hat):
        """
        WIP.

        Args:
            self, sigma
        Returns:
            np.array: GMM error
        Raises:
            -
        """
        params = np.concatenate((beta, sigma), axis=0)
        G_hat = self.__moment_jacobian(params)
        G_hat = G_hat[:, self.beta_len:]
        gradient = 2 * G_hat.T @ self.W @ g_hat
        return gradient
    
    
    def __calc_hausman_instruments(self):
        """
        Calculates Hausman-instruments. WIP.
        
        Args:
            self
        Returns:
            -
        Raises:
            -
        """
        prices = self.data['price'].to_numpy()
        instruments = np.zeros(shape=(self.N, self.T-1))
        product_counter = 0
        for product, indices in self.product_indices.items():
            for index in indices:
                prices_in_other_markets = prices[np.setdiff1d(indices, index)]
                instruments[index, :] = prices_in_other_markets.T
        self.Z_d = instruments
        
        

        