# -*- coding: utf-8 -*-
"""
Created on Mon Nov 20 10:47:27 2023

@author: Antto
"""

import os
import sys
import numpy as np
import pandas as pd
from patsy import dmatrix
import pyblp


# =============================================================================
#                               Globals
# =============================================================================

# Define root
root = r"C:\Users\Antto\Dropbox (Aalto)\Courses\Structural econometrics\Replication"

# Monte carlo iterations
I = 1000

# Random seed
seed = 0

# Utility function formulations (0 = constant excluded, 1 = constant included)
X1_formulation = '0 + prices + C(product_ids)' # linear variables
X2_formulation = '1 + prices + sugar + mushy' # non-linear variables

# Number of random coefficients
X2_len = len(X2_formulation.split('+'))


# =============================================================================
#                               Load data
# =============================================================================

# From PyBLP package - same as in Eric Rasmussen's website
product_data = pd.read_csv(pyblp.data.NEVO_PRODUCTS_LOCATION)



# =============================================================================
#                           PyBLP implementation 
# =============================================================================

# PyBLP settings
pyblp.options.digits = 2
pyblp.options.verbose = False

# Formulations
product_formulations = (pyblp.Formulation(X1_formulation), 
                        pyblp.Formulation(X2_formulation))

# Heterogeneity draws
mc_integration = pyblp.Integration('monte_carlo', 
                                   size=I, 
                                   specification_options={'seed': seed})

# Optimizer
bfgs = pyblp.Optimization('l-bfgs-b', {'gtol': 1e-6})

# Initialize estimator
mc_problem = pyblp.Problem(product_formulations, 
                           product_data, 
                           integration=mc_integration)

# Initial guesses for sigmas (independent RCs)
init_sigma = sigma=np.eye(X2_len)

# Estimate
results1 = mc_problem.solve(init_sigma, 
                            optimization=bfgs, 
                            method='2s', 
                            finite_differences=True)


# =============================================================================
#                           Our implementation
# =============================================================================

# Load our demand estimation package
sys.path.append(root)
from demand_estimation import DemandEstimation


# Instruments
filter_col = [col for col in product_data if col.startswith('demand_instruments')]
Z_d = product_data[filter_col] # demand side instruments

# Initial guess for sigmas
init_sigma = np.ones(shape=(X2_len,))

# Initialize estimator
estimator = DemandEstimation(product_data, 
                             X1_formulation,
                             X2_formulation, 
                             Z_d,
                             init_sigma,
                             I,
                             seed)

# Estimate
estimator.estimate(twostep=True)
results2 = estimator



# =============================================================================
#                               Markups
# =============================================================================


# Prices
prices = product_data['prices'].to_numpy()[:, np.newaxis]

# Init vectors of markups
markups_our_model = []
markups_pyblp = []


### Independent firms ###

# Calculate ownership matrices as np.arrays within a dictionary
ownership = {}
for market, indices in results2.market_indices.items():
    ownership[market] = np.eye(len(indices))
    
# Calculate median markup for our model
marginal_costs = results2.marginal_costs(ownership)
markups = (prices - marginal_costs)/prices
markups_our_model.append(np.median(markups)*100)

# Stack the matrices within the dictionary and solve pyblp median markup
ownership = np.concatenate([matrix for matrix in ownership.values()], axis=0)
marginal_costs = results1.compute_costs(ownership = ownership)
markups = (prices - marginal_costs)/prices
markups_pyblp.append(np.median(markups)*100)


### Current ownership ###

# Calculate ownership matrix as before
ownership = {}
for market, indices in results2.market_indices.items():
    firms_t = results2.data.loc[indices, 'firm_ids'].to_numpy()
    ownership_t = np.empty(shape=(len(indices), len(indices)))
    for i in range(len(indices)):
        for j in range(len(indices)):
            ownership_t[i, j] = (firms_t[i] == firms_t[j])
    ownership[market] = ownership_t
    
# Calculate median markup for our model
marginal_costs = results2.marginal_costs(ownership)
markups = (prices - marginal_costs)/prices
markups_our_model.append(np.median(markups)*100)

# Solve pyblp median markup (uses firm_id as default)
marginal_costs = results1.compute_costs()
markups = (prices - marginal_costs)/prices
markups_pyblp.append(np.median(markups)*100)


### Monopolist ###

# Calculate ownership matrix as before
ownership = {}
for market, indices in results2.market_indices.items():
    ownership[market] = np.ones(shape=(len(indices), len(indices)))
    
# Calculate median markup for our model
marginal_costs = results2.marginal_costs(ownership)
markups = (prices - marginal_costs)/prices
markups_our_model.append(np.median(markups)*100)

# Stack the matrices within the dictionary and solve pyblp median markup
ownership = np.concatenate([matrix for matrix in ownership.values()], axis=0)
marginal_costs = results1.compute_costs(ownership = ownership)
markups = (prices - marginal_costs)/prices
markups_pyblp.append(np.median(markups)*100)




# =============================================================================
#                           Print results
# =============================================================================


### Nevo's results ###

nevo_beta = [-27.198]
nevo_beta_se = [5.248]

nevo_sigma = [0.330, 2.453, 1.661, 0.244]
nevo_sigma_se = [0.609, 2.978, 5.866, 0.623]

markups_nevo = [35.8, 42.2, 72,6]



### Print betas and sigmas ###

print("PyBLP:")
print("Alpha:")
print("Mean:", results1.beta[0,0])
print("Std:", results1.beta_se[0,0])
print("Sigmas:")
print("Mean:", np.diagonal(results1.sigma))
print("Std:", np.diagonal(results1.sigma_se))
print("--------------------")

print("Our model:")
print("Alpha:")
print("Mean:", results2.beta[-1])
print("Std:", results2.beta_se[-1])
print("Sigmas:")
print("Mean:", results2.sigma)
print("Std:", results2.sigma_se)
print("--------------------")


output_file = os.path.join(root, "estimation_table.tex")

with open(output_file, 'w') as f:
    
    rows = [r"\centering", "\n",
            r"\begin{tabular}{l c c c}", "\n",
            r"\hline\hline", "\n",
            r" & Our model & PyBLP & Nevo\\", "\n",
            r"\hline", "\n",
            fr" {results2.X1_labels[-1]} & {results2.beta[-1]:.3f} & {results1.beta[0,0]:.3f} & {nevo_beta[0]:.3f} \\", "\n"
            fr" & ({results2.beta_se[-1]:.3f}) & ({results1.beta_se[0,0]:.3f}) & ({nevo_beta_se[0]:.3f}) \\", "\n"]
    f.writelines(rows)
    for i in range(X2_len):
       f.write(fr" RC: {results2.X2_labels[i]} & {results2.sigma[i]:.3f} & {np.diagonal(results1.sigma)[i]:.3f} & {nevo_sigma[i]:.3f} \\")
       f.write("\n")
       f.write(fr"  & ({results2.sigma_se[i]:.3f}) & ({np.diagonal(results1.sigma_se)[i]:.3f}) & ({nevo_sigma_se[i]:.3f}) \\")
       f.write("\n")
    f.write(r"\hline\hline")
    f.write("\n")
    f.write(r"\end{tabular}")
    f.write("\n")


### Print markups ###

print("PyBLP")
print("Markup-% (Independent):")
print("Median:", markups_pyblp[0])

print("Markup-% (Current ownership):")
print("Median:", markups_pyblp[1])

print("Markup-% (Monopolist):")
print("Median:", markups_pyblp[2])

print("Our model")
print("Markup-% (Independent):")
print("Median:", markups_our_model[0])

print("Markup-% (Current ownership):")
print("Median:", markups_our_model[1])

print("Markup-% (Monopolist):")
print("Median:", markups_our_model[2])


output_file = os.path.join(root, "markups_table.tex")

with open(output_file, 'w') as f:
    
    rows = [r"\centering", "\n",
            r"\begin{tabular}{l c c c}", "\n",
            r"\hline\hline", "\n"
            r" & Our model & PyBLP & Nevo \\", "\n",
            r"\hline", "\n",
            fr" Independent & {markups_our_model[0]:.1f}\% & {markups_pyblp[0]:.1f}\% & {markups_nevo[0]:.1f}\% \\", "\n",
            fr" Current ownership & {markups_our_model[1]:.1f}\% & {markups_pyblp[1]:.1f}\% & {markups_nevo[1]:.1f}\% \\", "\n",
            fr" Monopolist & {markups_our_model[2]:.1f}\% & {markups_pyblp[2]:.1f}\% & {markups_nevo[2]:.1f}\% \\", "\n",
            r"\hline\hline", "\n",
            r"\end{tabular}"]
    
    f.writelines(rows)

